import functools
import h5py
from qtpy import QtCore, QtWidgets

from ScopeFoundry.data_browser.data_browser_plug_in import DataBrowserPlugIn


class H5SearchPlugIn(DataBrowserPlugIn):
    name = "h5_search"
    button_text = "🔍h5"
    show_keyboard_key = QtCore.Qt.Key_F
    description = "used to inspect data sets and attributes of .h5 files (Ctrl+F)"

    def setup(self):
        self.search_line = QtWidgets.QLineEdit()
        self.search_line.setText("measurement")
        self.text_edit = QtWidgets.QTextEdit()

        search_layout = QtWidgets.QHBoxLayout()
        search_layout.addWidget(QtWidgets.QLabel("🔍 text"))
        search_layout.addWidget(self.search_line)

        self.ui = QtWidgets.QWidget(objectName="SearchWidget")
        self.ui.setMaximumHeight(1200)

        layout = QtWidgets.QVBoxLayout(self.ui)
        layout.addLayout(search_layout)
        layout.addWidget(self.text_edit)
        self.ui.setStyleSheet(
            "QWidget#SearchWidget{background-color:rgba(0, 166, 237, 0.1)}"
        )

        self.search_line.textChanged.connect(self.update)

    def update(self, fname: str = None) -> None:
        fname = self.new_fname
        if not fname.endswith(".h5"):
            self.text_edit.setText(f"{fname} is not supported")
            return
        search_text = self.search_line.text()
        self.new_search(search_text, fname)

    def new_search(self, search_text, fname):
        # x = search_text.lower()
        if search_text == "":
            text = "<br>".join(make_tree(fname))
        else:
            text = "<br>".join(search_h5(fname, search_text))
            text = text.replace(
                search_text,
                f"<font color='green'>{search_text}</font>",
            )
        self.text_edit.setText(text)


def search_h5(fname, search_text):
    priority_results = []
    results = []
    visit_func = functools.partial(
        _search_visitfunc,
        results=results,
        priority_results=priority_results,
        search_text=search_text,
    )
    with h5py.File(fname) as file:
        file.visititems(visit_func)
    return priority_results + results


def _search_visitfunc(name, node, results, priority_results, search_text):
    if isinstance(node, h5py.Dataset):
        if not search_text in name:
            return
        try:
            vals = node[:].ravel()
            if len(vals) < 4:
                stats = str(vals)
            else:
                stats = f"min={min(vals):1.1f} max={max(vals):1.1f}"
        except ValueError:
            stats = ""
        res = f"<i>{name}, {node.shape}, {node.dtype}</i> {stats}"

        if search_text in res:
            priority_results.append(res)
        elif search_text.lower() in res.lower():
            results.append(res)

    if name.endswith("settings"):
        for key, val in node.attrs.items():
            units = node["units"].attrs[key] if key in node["units"].attrs else ""
            res = f"<b>{name.replace('settings', key)}</b>: {str(val)} {units}"
            if search_text in res:
                priority_results.append(res)
            elif search_text.lower() in res.lower():
                results.append(res)


def make_tree(fname):
    texts = []
    visit_func = functools.partial(_tree_visitfunc, texts=texts)
    with h5py.File(fname) as file:
        file.visititems(visit_func)
    return texts


def _tree_visitfunc(name, node, texts):
    level = len(name.split("/"))
    indent = "&nbsp;" * 4 * (level - 1)
    localname = name.split("/")[-1]

    if isinstance(node, h5py.Group):
        t = f"|> <b>{localname}/</b>".format(localname)
    elif isinstance(node, h5py.Dataset):
        t = f"|D <b>{localname}</b>: {node.shape} {node.dtype}"
    texts.append(indent + t)

    for key, val in node.attrs.items():
        t = f"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|- <i>{key}</i> = {val}"
        texts.append(indent + t)



from qtpy.QtWidgets import (QCheckBox, QComboBox, QGroupBox, QHBoxLayout,
                            QLabel, QLineEdit, QPushButton, QTextEdit,
                            QVBoxLayout)


class ImportFromGhView(QGroupBox):

    def __init__(self) -> None:
        super().__init__()
        self.setup()

    def setup(self) -> None:
        layout = QVBoxLayout()
        self.setLayout(layout)

        self.origin_repo_label = QLabel('origin repo')
        self.origin_repo_le = QLineEdit(
            'https://github.com/ScopeFoundry/HW_acton_spec.git')
        self.origin_repo_validat_btn = QPushButton('validate')
        self.origin_repo_validat_btn.setMaximumWidth(55)
        self.import_btn = QPushButton('Import from origin')
        self.import_btn.setEnabled(False)
        origin_layout = QHBoxLayout()
        origin_layout.addWidget(self.origin_repo_label)
        origin_layout.addWidget(self.origin_repo_le)
        origin_layout.addWidget(self.origin_repo_validat_btn)
        origin_layout.addWidget(self.import_btn)
        layout.addLayout(origin_layout)

        self.my_gh_acc_label = QLabel('my GitHub')
        self.my_gh_acc_le = QLineEdit('UBene')
        self.my_gh_acc_validate_btn = QPushButton('validate')
        self.my_gh_acc_validate_btn.setMaximumWidth(55)
        self.my_gh_acc_is_valid_cb = QCheckBox('')
        self.my_gh_acc_is_valid_cb.setEnabled(False)
        self.my_gh_acc_is_valid_cb.setMaximumWidth(15)
        self.fork_btn = QPushButton('Fork to my gh')
        self.validate_fork_btn = QPushButton('validate fork')
        self.import_from_fork_btn = QPushButton('import from fork')
        self.import_from_fork_btn.setEnabled(False)
        fork_layout = QHBoxLayout()
        fork_layout.addWidget(self.my_gh_acc_label)
        fork_layout.addWidget(self.my_gh_acc_le)
        fork_layout.addWidget(self.fork_btn)
        fork_layout.addWidget(self.validate_fork_btn)
        fork_layout.addWidget(self.import_from_fork_btn)
        layout.addLayout(fork_layout)

        self.info_box = QTextEdit('')
        layout.addWidget(self.info_box)

    def validate_origin_repo_clicked(self, callback):
        self.origin_repo_validat_btn.clicked.connect(callback)

    def validate_my_gh_acc_validate_clicked(self, callback):
        self.my_gh_acc_validate_btn.clicked.connect(callback)

    def fork_clicked(self, callback):
        self.fork_btn.clicked.connect(callback)

    def validate_fork_clicked(self, callback):
        self.validate_fork_btn.clicked.connect(callback)

    def import_from_fork_clicked(self, callback):
        self.import_from_fork_btn.clicked.connect(callback)

    def set_ready_to_import_from_fork(self, ready):
        self.import_from_fork_btn.setEnabled(ready)

    def import_clicked(self, callback):
        self.import_btn.clicked.connect(callback)

    def set_ready_to_import(self, ready):
        self.import_btn.setEnabled(ready)

    def set_my_gh_acc_valid(self, checked):
        self.my_gh_acc_is_valid_cb.setChecked(checked)

    def get_origin_repo(self) -> str:
        return self.origin_repo_le.text()

    def set_origin_repo(self, text):
        self.origin_repo_le.setText(text)

    def get_my_gh_acc(self) -> str:
        return self.my_gh_acc_le.text()

    def set_my_gh_acc(self, text):
        self.my_gh_acc_le.setText(text)

    def set_info_text(self, text):
        self.info_box.setHtml(text)

    def append_info(self, text):
        self.info_box.setHtml(self.info_box.toHtml() + text)



'''
Created on $DATE_PRETTY

@author: $AUTHORS
'''
from ScopeFoundry.hardware import HardwareComponent


class $HW_CLASS_NAME(HardwareComponent):

    name = "$HW_NAME"

    def setup(self):


        self.settings.New('dll_path',
                          str,
                          initial=r"path_to_dll")

        self.settings.New("property_x",
                          dtype=float,
                          ro=False,
                          unit='mm',
                          spinbox_decimals=4,
                          si=False
                          )
                          
    def connect(self):
        $IMPORT_DEV

        S = self.settings

        self.dev = $DEV_CLASS_NAME(
            dll_path=S['dll_path'],
            debug=S['debug_mode'])

        # attache read and write functions to settings
        S.property_x.connect_to_hardware(
            self.dev.read_property_x,
            self.dev.write_property_x)

    def disconnect(self):
        if not hasattr(self, 'dev'):
            return

        # self.settings.disconnect_all_from_hardware()
        self.dev.close()
        del self.dev

    # if you want to continuously update settings implement *run* method
    # def run(self):
    #     self.settings.property_x.read_from_hardware()



"""
Created on Sep 17, 2021

@author: Benedikt Ursprung
"""
import glob
import json
import os
import time
from builtins import getattr
from typing import List, Tuple, Dict

from qtpy.QtCore import Qt
from qtpy import QtWidgets


from ScopeFoundry import Measurement
from .item_types.editor_base_controller import EditorBaseController
from .item_types.editor_base_ui import EditorBaseUI
from .item_types.item_list import ItemList
from .item_types import (
    BaseItem,
    ExecFunctionEditorUI,
    InterationsEditorController,
    IterationsEditorUI,
    IterruptIfEditorUI,
    NewDirEditorUI,
    PauseEditorUI,
    ReadFromHardWareEditorUI,
    RunMeasurementEditorUI,
    SaveDirToParentEditorUI,
    TimeoutEditorUI,
    UpdateSettingEditorUI,
    VisitReturnType,
    WaitUntilEditorUI,
    link_iteration_items,
    new_item,
)


class Sequencer(Measurement):
    name = "sequencer"

    def setup(self):
        self.settings.New(
            "cycles",
            int,
            initial=1,
            description="number of times the sequence is executed",
        )
        self.settings.New("paused", bool, initial=False)
        self.iter_values = {}
        self.letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        self.current_item = None

    def setup_figure(self):
        self.ui = QtWidgets.QWidget()
        self.layout = QtWidgets.QVBoxLayout(self.ui)

        # measurement controls and settings
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.settings.get_lq("cycles").new_default_widget())
        layout.addWidget(self.settings.get_lq("activation").new_pushButton())
        btn = self.settings.get_lq("paused").new_pushButton(
            texts=["pause", "resume"], colors=[None, "rgba( 0, 255, 0, 220)"]
        )
        layout.addWidget(btn)
        self.layout.addLayout(layout)

        # select file combobox
        self.load_file_comboBox = QtWidgets.QComboBox()
        self.update_load_file_comboBox()
        self.load_file_comboBox.currentTextChanged.connect(
            self.on_load_file_comboBox_changed
        )
        self.layout.addWidget(self.load_file_comboBox)

        # item list
        self.item_list = ItemList()
        self.layout.addWidget(self.item_list.get_view())
        self.item_list.connect_item_double_clicked(self.item_double_clicked)

        # controls
        layout = QtWidgets.QHBoxLayout()
        self.layout.addLayout(layout)
        self.remove_pushButton = QtWidgets.QPushButton("remove selected item")
        self.remove_pushButton.setToolTip("DEL")
        self.remove_pushButton.clicked.connect(self.on_remove_item)
        layout.addWidget(self.remove_pushButton)
        btn = QtWidgets.QPushButton("save list ...")
        btn.clicked.connect(self.on_save)
        layout.addWidget(btn)
        btn = QtWidgets.QPushButton("load list ...")
        btn.clicked.connect(self.on_load)
        layout.addWidget(btn)
        btn = QtWidgets.QPushButton("run selected item")
        btn.setToolTip("SPACEBAR")
        btn.clicked.connect(self.on_run_item_and_proceed)
        layout.addWidget(btn)
        self.show_editor_checkBox = QtWidgets.QCheckBox("show|hide editor")
        layout.addWidget(self.show_editor_checkBox)

        # Editors
        self.editor_widget = QtWidgets.QWidget()
        self.editor_layout = QtWidgets.QVBoxLayout()
        self.editor_widget.setLayout(self.editor_layout)
        self.layout.addWidget(self.editor_widget)
        self.show_editor_checkBox.stateChanged.connect(self.editor_widget.setVisible)
        self.show_editor_checkBox.setChecked(True)

        paths = self.app.get_setting_paths()
        funcs = get_all_functions(self.app)
        self.editors: Dict[str, EditorBaseController] = {}
        self.add_editor(ReadFromHardWareEditorUI(self, paths))
        self.add_editor(UpdateSettingEditorUI(self, paths))
        self.add_editor(RunMeasurementEditorUI(self))
        self.add_editor(WaitUntilEditorUI(self, paths))
        self.add_editor(WaitUntilEditorUI(self, paths))
        self.add_editor(ExecFunctionEditorUI(self, funcs))
        self.add_editor(PauseEditorUI(self))
        self.add_editor(IterruptIfEditorUI(self, paths))
        self.add_editor(NewDirEditorUI(self))
        self.add_editor(SaveDirToParentEditorUI(self))
        self.add_editor(TimeoutEditorUI(self))
        self.add_iteration_editor(IterationsEditorUI(self, paths))

        for editor in self.editors.values():
            self.editor_layout.addWidget(editor.ui.group_box)

        self.editor_widget.keyPressEvent = self._editorKeyPressEvent
        self.item_list.get_view().keyReleaseEvent = self._keyReleaseEvent

    def add_editor(self, editor_ui: EditorBaseUI):
        self.editors[editor_ui.item_type] = EditorBaseController(editor_ui, self)

    def add_iteration_editor(self, editor_ui: IterationsEditorUI):
        editor = InterationsEditorController(editor_ui, self)
        self.editors["start-iteration"] = editor
        self.editors["end-iteration"] = editor

    def _editorKeyPressEvent(self, event):
        if not event.modifiers() & Qt.ControlModifier:
            return
        if not event.key() in (Qt.Key_R, Qt.Key_N):
            return
        fw = self.editor_widget.focusWidget()
        # find editor with focused widget
        for e in self.editors.values():
            gb = e.ui.group_box
            if fw in gb.findChildren(type(fw), fw.objectName()):
                if event.key() == Qt.Key_R:
                    e.on_replace_func()
                if event.key() == Qt.Key_N:
                    e.on_new_func()

    def _keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Delete:
            self.item_list.remove()
        if event.key() == Qt.Key_Space:
            self.on_run_item_and_proceed()
        if event.key() in (Qt.Key_Enter, Qt.Key_Return):
            item = self.item_list.get_current_item()
            self.item_double_clicked(item)

    def update_load_file_comboBox(self):
        fnames = glob.glob(glob.os.getcwd() + "\..\..\**/*.json", recursive=True)
        index0 = self.load_file_comboBox.currentIndex()
        self.load_file_comboBox.clear()
        self.seq_fnames = {}
        for fname in fnames:
            abbrev_fname = "\\".join(fname.split("\\")[-2:])
            self.seq_fnames.update({abbrev_fname: fname})
        self.load_file_comboBox.addItems(list(self.seq_fnames.keys()))
        self.load_file_comboBox.setCurrentIndex(index0)

    def on_load_file_comboBox_changed(self, fname):
        self.load_file(self.seq_fnames[fname])

    def next_iter_id(self) -> str:
        return self.letters[self.item_list.count_type("start-iteration")]

    def on_remove_item(self):
        self.item_list.remove(None)

    def on_save(self) -> str:
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self.ui, caption="Save Sequence", filter="Sequence (*.json)"
        )
        if fname:
            self.save_to_file(fname)
        self.update_load_file_comboBox()
        return fname

    def save_to_file(self, fname):
        with open(fname, "w+") as f:
            f.write(json.dumps(self.item_list.as_dicts(), indent=1))

    def on_load(self) -> str:
        fname, _ = QtWidgets.QFileDialog.getOpenFileName(
            None, filter="Sequence (*.json)"
        )
        if fname:
            self.load_file(fname)
        return fname

    def load_file(self, fname):
        self.item_list.clear()
        with open(fname, "r") as f:
            lines = json.loads(f.read())
        for kwargs in lines:
            item_type = kwargs.pop("type")
            item = new_item(self, item_type, **kwargs)
            self.item_list.add(item)
        success = link_iteration_items(self.item_list)
        if not success:
            print("invalid list")

    def on_run_item(self) -> Tuple[BaseItem, VisitReturnType]:
        item = self.item_list.get_current_item()
        if item.item_type == "measurement":
            print("WARNING running individually measurement not supported")
            return (item, None)
        return (item, self.item_list.get_current_item().visit())

    def on_run_item_and_proceed(self):
        item, next_item = self.on_run_item()
        if next_item is None:
            row = self.item_list.get_row(item)
            next_item = self.item_list.get_item(row + 1)
        self.item_list.set_current_item(next_item)

    def item_double_clicked(self, item: BaseItem):
        self.editors[item.item_type].ui.set_kwargs(**item.kwargs)

    def run(self):
        success = link_iteration_items(self.item_list)
        if not success:
            print("invalid list")

        N = self.item_list.count()
        for i in range(N):
            self.item_list.get_item(i).reset()

        for q in range(self.settings["cycles"]):
            # pct = int(100 * q / self.settings['cycles'])
            # self.set_progress(pct)
            if self.interrupt_measurement_called:
                break

            # go through list
            row = 0
            while row < self.item_list.count():
                while self.settings["paused"]:
                    if self.interrupt_measurement_called:
                        break
                    time.sleep(0.03)

                self.current_item = item = self.item_list.get_item(row)

                resp = item.visit()
                if resp is None:
                    # go to next item
                    row += 1
                else:
                    # jump to item returned
                    row = self.item_list.get_row(resp)

                if self.interrupt_measurement_called:
                    break

        self.current_item = None

    def update_display(self):
        for i in range(self.item_list.count()):
            item = self.item_list.get_item(i)
            if item == self.current_item:
                item.setBackground(Qt.green)
            else:
                item.setBackground(Qt.white)

    def shutdown(self):
        os.system("shutdown /s /f /t 1")

    def restart(self):
        os.system("restart /r /f /t 1")


def get_all_functions(app) -> List[str]:
    funcs = []

    def append_objs_callables(obj, from_app_path):
        for a in dir(obj):
            try:  # Not sure why some python version seem to need this block
                if callable(getattr(obj, a)) and not a.startswith("__"):
                    funcs.append(f"{from_app_path}{obj.name}.{a}")
            except AttributeError as e:
                print(e)

    append_objs_callables(app, "")
    for m in app.measurements.values():
        append_objs_callables(m, "measurements.")
    for h in app.hardware.values():
        append_objs_callables(h, "hardware.")
    return funcs



from ScopeFoundry import LQCollection, BaseApp

class LQConnectionTestApp(BaseApp):
    
    name = 'LQConnectionTestApp'

    def __init__(self,argv):
        BaseApp.__init__(self,argv)
        
        lq0 = self.settings.New('lq0', dtype=float,ro=False, initial=5)
        lq1 = self.settings.New('lq1', dtype=float,ro=False, initial=5)
        lq2 = self.settings.New('lq2', dtype=float,ro=False, initial=35)
        lq3 = self.settings.New('lq3', dtype=float,ro=True, initial=35)
        lq4 = self.settings.New('lq4', dtype=float,ro=False, initial=35)
        lq_sum = self.settings.New('lq_sum', dtype=float,ro=True, initial=35)
        lq_sum_bidir = self.settings.New('lq_sum_bidir', dtype=float,ro=False, initial=35)
        
        lq_scale = self.settings.New('lq_scale', dtype=float, ro=False)
        lq_scale2 = self.settings.New('lq_scale2', dtype=float, ro=False)
        
        
        lq_array = self.settings.New('lq_array', dtype=float, array=True, initial=[1.0,2.0,3.0,4.0])
        lq_array_element = self.settings.New('lq_array_element', dtype=float)
        
        lq_array_element.connect_lq_math( (lq_array,), lambda arr: arr[1])
        

        lq1.connect_to_lq(lq2)
        
        lq3.connect_lq_math(lq1, lambda x1: x1 + 1 )
        

        lq4.connect_lq_math(lq1,
                     func=lambda x: x + 5,
                     reverse_func= lambda y: y - 5, 
                     )


        def sum_lq(*vals):
            return sum(vals)

        lq_sum.connect_lq_math((lq1,lq2, lq3, lq4),
                        func=lambda a,b,c,d: a+b+c+d)
        
        
        
        def lq_sum_reverse(new_val, old_vals):
            a,b = old_vals
            return new_val - b, b
        lq_sum_bidir.connect_lq_math((lq0,lq1),
                              func=sum_lq,
                              reverse_func=lq_sum_reverse)
        
        
        lq_scale.connect_lq_math((lq0,), func=lambda x: 10*x,
                          reverse_func=lambda y, : [0.1*y,])
        
        lq_scale2.connect_lq_scale(lq0, 25.0)
        
        
        
        test_array = self.settings.New('test_array', dtype=float, array=True,  ro=False, fmt="%1.2f",
          initial=[[147, 111 , 100]])

        array_follower = self.settings.New('array_follower', dtype=float)
        
        test_array.connect_element_follower_lq( array_follower, index=(0,1), bidir=True )
        
        
        
        self.ui = self.settings.New_UI()
        
        self.console_widget.show()
        self.ui.show()
        
        
if __name__ == '__main__':
    app = LQConnectionTestApp([])
    app.exec_()


import uuid
import base64 # part of standard library

b32_alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567"
cb32_alphabet = "0123456789abcdefghjkmnpqrstvwxyz"

b32_to_cb32_map = dict(zip(b32_alphabet,cb32_alphabet))

def b32_to_cb32(x):
    """
    Map a base32 encoded string (without padding) 
    to a Crockford's Base32 encoded string
    
    https://www.crockford.com/base32.html
    """
    return "".join(
        [ b32_to_cb32_map[k] for k in x ]
    )
    
def uuid_to_cb32(u):
    """Creates a 26 character encoded string based on lowercase
    Crockford's Base32 encoding of the uuid.bytes
    
    u: uuid object
    returns string of Crockford's Base32
    """
    # note we use typical base32 RFC 4648 alphabet in the base64 
    # standard library and then convert to Crockford's by alphabet mapping
    b32 = base64.b32encode(u.bytes).decode().strip("=")
    cb32 = b32_to_cb32(b32)
    return cb32


def cb32_uuid4():
    """
    Creates a 26 character encoded string based on lowercase
    Crockford's Base32 (cb32) encoding of the uuid.bytes (UUIDv7) 

    Useful as a compact random, globally unique
    identifier that can be used as a filename on most operating systems
    
    returns cb32 string and UUIDv4 object
    """
    u = uuid.uuid4()
    cb32 = uuid_to_cb32(u)
    return cb32,u

def cb32_uuid7(*args, **kwargs):
    """
    Creates a 26 character encoded string based on lowercase
    Crockford's Base32 (cb32) encoding of the uuid.bytes (UUIDv7) 

    Useful as a compact time-based lexicographically-sortable
    identifier that can be used as a filename on most operating systems

    args and kwargs are passed to uuid7()
    
    returns cb32 string and UUIDv7 object
    """
    from uuid_extensions import uuid7 # pip install uuid7

    u = uuid7(*args, **kwargs)
    cb32 = uuid_to_cb32(u)
    return cb32,u

def cb32_uuid():
    """
    Creates a 26 character encoded string based on lowercase
    Crockford's Base32 encoding of a UUID. 
    Uses a time sequential UUIDv7 if available,
    otherwise create a random UUIDv4
    
    returns cb32 string and UUID object
    """    
    try:
        return cb32_uuid7()
    except:
        return cb32_uuid4()
    
if __name__ == '__main__':
    print(cb32_uuid())